# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Provisional backward compatible version of prior client for to BlueCat Address Manager REST API.

.. versionadded:: 20.12.1
"""
from bluecat_libraries.address_manager.api.rest.auto import _wadl_parser
from bluecat_libraries.address_manager.api.rest.auto.client import _get_rest_service_base_url
from bluecat_libraries.address_manager.api.rest.provisional.rest_dict import RESTDict
from bluecat_libraries.address_manager.api.rest.provisional.rest_entity_array import RESTEntityArray
from bluecat_libraries.address_manager.api.rest.provisional.rest_fault import RESTFault
from bluecat_libraries.http_client import GeneralError, ErrorResponse
import requests


class ProvisionalClient:
    """
    A client to BlueCat Address Manager REST API. It is to be used during
    the transitional period while we migrate away from the `SOAP` and
    `suds`-influenced legacy designs.

    .. versionadded:: 20.12.1

    .. versionchanged:: 21.5.1
        The initializer of the class no longer takes in a ``url`` and ``kwargs``, but an instance
        of a raw client.
    """

    def __init__(self, raw_client):
        self._raw_client = raw_client

    def login(self, username, password):
        """
        Wrapper for the REST login API/endpoint.

        :param username: Username of the user that logs in.
        :type username: str
        :param password: Password of the user that logs in.
        :type password: str
        :return: Access token for BlueCat Address Manager.
        :rtype: str
        """
        try:
            return self._raw_client.login(username, password)
        except ErrorResponse as exc:
            raise RESTFault(exc.response.text, exc.response) from exc
        except GeneralError as exc:
            raise RESTFault(str(exc)) from exc

    def loginWithOptions(self, username, password, options):
        """
        Wrapper for the REST login API/endpoint.

        :param username: Username of the user that logs in.
        :type username: str
        :param password: Password of the user that logs in.
        :type password: str
        :param options: Login options.
        :type options: str
        :return: Access token for BlueCat Address Manager.
        :rtype: str
        """
        try:
            return self._raw_client.loginWithOptions(username, password, options)
        except ErrorResponse as e:
            raise RESTFault(e.response.text, e.response) from e
        except GeneralError as e:
            raise RESTFault(str(e)) from e

    @classmethod
    def process_url(cls, url, **kwargs):  # pylint: disable=unused-argument
        """
        Returns the base URL to the BAM's REST API service based on an URL to a BAM server.

        :param url: URL to a BAM server.
        :type url: str
        :param kwargs: Various options.
        :type kwargs: dict
        :return: The URL to the base path for the REST service of BAM.
        :rtype: str

        .. note::

           This is a provisionally provided method to ensure backward compatibility. It will be
           deprecated and removed in a future release.

        .. versionadded:: 20.12.1
        """
        # NOTE: Delay raising deprecation warning until Gateway's own code will not trigger it.
        # warnings.warn(
        #     "Class method 'Client.process_url' has been deprecated in favour of"
        #     " function 'bluecat_libraries.address_manager.api.rest.auto.get_rest_service_base_url'."
        #     " Be advised that the parameters have changed.",
        #     DeprecationWarning)
        return _get_rest_service_base_url(url, use_https=url.startswith("https"))

    @property
    def service(self) -> object:
        """
        Provides access to the internal service with automatically generated methods for the
        endpoints of the BAM REST API.

        .. note::

           This is a provisionally provided property to ensure backward compatibility. It will be
           deprecated and removed in a future release.

        .. versionadded:: 20.12.1
        """
        # NOTE: Delay raising deprecation warning until Gateway's own code will not trigger it.
        # warnings.warn(
        #     "Member 'service' has been deprecated and will be removed in a future release."
        #     " You may now use the instance of 'Client' directly the same way you would have used 'service'.",
        #     DeprecationWarning)
        return self

    @property
    def session(self) -> requests.Session:
        """
        Provides access to the internal Session instance used for HTTP communication.

        .. note::

           This is a provisionally provided property to ensure backward compatibility. It will be
           deprecated and removed in a future release.

        .. versionadded:: 21.5.1
        """
        return self._raw_client.session

    @property
    def url(self) -> str:
        """
        Provides the base URL of the endpoints of the BAM REST API service.

        .. note::

           This is a provisionally provided property to ensure backward compatibility. It will be
           deprecated and removed in a future release.

        .. versionadded:: 20.12.1
        """
        # NOTE: Delay raising deprecation warning until Gateway's own code will not trigger it.
        # warnings.warn(
        #     "Member 'url' has been deprecated and will be removed in a future release."
        #     " It is available for a limited transitional period.",
        #     DeprecationWarning)
        return self._service.url.rstrip("/")

    @classmethod
    def _get_namespace(cls, element):
        # NOTE: Delay raising deprecation warning until Gateway's own code will not trigger it.
        # warnings.warn(
        #     "Class method '_get_namespace' has been deprecated in favour of"
        #     " function 'bluecat_libraries.address_manager.api.rest.auto.wadl_parser.get_element_namespace'.",
        #     DeprecationWarning)
        return _wadl_parser.get_element_namespace(element)

    def __getattr__(self, item):
        # Obtain the actual method from the parent class. Then capture its
        # output or exceptions and wrap them in the currently expected envelopes
        # for result or error.
        func = getattr(self._raw_client, item)

        def func_wrapper(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
                if isinstance(result, list):
                    return RESTEntityArray(result)
                if isinstance(result, dict):
                    return RESTDict(result)
                return result
            except ErrorResponse as exc:
                raise RESTFault(exc.response.text, exc.response) from exc
            except GeneralError as exc:
                raise RESTFault(str(exc)) from exc

        return func_wrapper



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Definition of Service Point as a target instance."""
from bluecat_libraries.http_client.instance import Instance

__all__ = [
    "ServicePointInstance",
]


class ServicePointInstance(Instance):
    """
    Definition of Service Point as a target instance.

    .. versionadded:: 20.6.1
    """

    def parse_url(self):
        """
        Process the service's URL and construct the value of the base URL of
        the service's API.
        """
        self.api_url_base = self.url._replace(path="/")



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Types of discovery methods to use for the network discovery operation in BlueCat Address Manager.
"""
from ._enum import StrEnum


class DiscoveryType(StrEnum):
    """Constants used for network discovery types."""

    NO_DISCOVERY = "NO_DISCOVERY"
    PINGSWEEP = "PINGSWEEP"
    SNMP = "SNMP"
    SNMP_PINGSWEEP = "SNMP_PINGSWEEP"



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Package for working with provisinal REST API"""
from bluecat_libraries.address_manager.api.rest.provisional.client import ProvisionalClient
from bluecat_libraries.address_manager.api.rest.provisional.rest_dict import RESTDict
from bluecat_libraries.address_manager.api.rest.provisional.rest_entity_array import RESTEntityArray
from bluecat_libraries.address_manager.api.rest.provisional.rest_fault import RESTFault

__all__ = [
    "ProvisionalClient",
    "RESTDict",
    "RESTEntityArray",
    "RESTFault",
]



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in methods related to DNS response policies in BlueCat Address Manager."""
from ._enum import StrEnum


class ResponsePolicy(StrEnum):
    """Constants used in to define the response policy type."""

    BLACKHOLE = "BLACKHOLE"
    BLACKLIST = "BLACKLIST"
    REDIRECT = "REDIRECT"
    WHITELIST = "WHITELIST"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in the DHCPv6 Deployment Option API methods in BlueCat Address Manager."""
from ._enum import StrEnum


class DHCP6ClientDeploymentOptionType(StrEnum):
    """Values for types of DHCPv6 client options"""

    DNS_SERVERS = "dns-servers"
    DOMAIN_SEARCH_LIST = "domain-search-list"
    INFORMATION_REFRESH_TIME = "information-refresh-time"
    SNTP_SERVERS = "sntp-servers"
    UNICAST = "unicast"
    WPAD_URL = "wpad-url"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in POST /v1/addServer method in BlueCat Address Manager."""
from ._enum import StrEnum


class ServerCapabilityProfiles(StrEnum):
    """Values for types of server profiles in BlueCat Address Manager."""

    ADONIS_800 = "ADONIS_800"
    ADONIS_1200 = "ADONIS_1200"
    ADONIS_1900 = "ADONIS_1900"
    ADONIS_1950 = "ADONIS_1950"
    ADONIS_XMB2 = "ADONIS_XMB2"
    ADONIS_XMB3 = "ADONIS_XMB3"
    AFILIAS_DNS_SERVER = "AFILIAS_DNS_SERVER"
    DNS_DHCP_SERVER_20 = "DNS_DHCP_SERVER_20"
    DNS_DHCP_SERVER_45 = "DNS_DHCP_SERVER_45"
    DNS_DHCP_SERVER_60 = "DNS_DHCP_SERVER_60"
    DNS_DHCP_SERVER_100 = "DNS_DHCP_SERVER_100"
    DNS_DHCP_SERVER_100_D = "DNS_DHCP_SERVER_100_D"
    DNS_DHCP_GEN4_7000 = "DNS_DHCP_GEN4_7000"
    DNS_DHCP_GEN4_5000 = "DNS_DHCP_GEN4_5000"
    DNS_DHCP_GEN4_4000 = "DNS_DHCP_GEN4_4000"
    DNS_DHCP_GEN4_2000 = "DNS_DHCP_GEN4_2000"
    OTHER_DNS_SERVER = "OTHER_DNS_SERVER"



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Module with exceptions that may occur while working with BlueCat DNS Edge."""
from requests import Response
from bluecat_libraries.http_client.exceptions import ErrorResponse

__all__ = ["EdgeErrorResponse"]


class EdgeErrorResponse(ErrorResponse):
    """
    Exception that represents the error HTTP response that was received, but
    additionally exposes the error code, reported by the Edge CI API.

    :param message: Text about the error that occurred.
    :type message: str
    :param response: The HTTP response resulting from a request.
    :type response: requests.Response
    :param code: A code identifying the problem.
    :type code: str
    """

    def __init__(self, message: str, response: Response, code: str, *args, **kwargs):
        super().__init__(message, response, *args, **kwargs)
        self.code = code



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in the DHCP Deployment Role API methods in BlueCat Address Manager."""
from ._enum import StrEnum


class DHCPDeploymentRoleType(StrEnum):
    """DHCP Deployment Role Type"""

    NONE = "NONE"
    MASTER = "MASTER"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in the DHCP Define Range API methods in BlueCat Address Manager."""
from ._enum import StrEnum


class DHCPDefineRange(StrEnum):
    """DHCP Define Range"""

    AUTOCREATE_BY_SIZE = "AUTOCREATE_BY_SIZE"
    OFFSET_AND_SIZE = "OFFSET_AND_SIZE"
    START_ADDRESS_AND_SIZE = "START_ADDRESS_AND_SIZE"
    OFFSET_AND_PERCENTAGE = "OFFSET_AND_PERCENTAGE"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Module with exceptions that may occur while working with BlueCat Address Manager's GraphQL API"""
from typing import Any, Optional

from requests import Response

from bluecat_libraries.http_client.exceptions import ErrorResponse


class GraphQLException(ErrorResponse):
    """
    Exception that represents the the errors received from the GraphQL API.

    If this is an HTTP 400+ error, the ``errors`` attribute will be ``None``. If this is a GraphQL
    error, the ``errors`` attribute will be the ``errors`` field of the GraphQL response body.

    The ``message`` attribute is meant to be readable for humans. Do not parse it programmatically;
    use the ``response`` and ``errors`` attributes instead.
    """

    def __init__(
        self,
        message: str,
        response: Response,
        errors: Optional[list[dict[str, Any]]],
        *args,
        **kwargs,
    ) -> None:
        super().__init__(message, response, *args, **kwargs)
        self.errors = errors



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Module for providing base classes for clients."""
from bluecat_libraries.http_client.client import Client
from bluecat_libraries.http_client.exceptions import (
    GeneralError,
    ClientError,
    ErrorResponse,
    UnexpectedResponse,
)
from bluecat_libraries.http_client.instance import Instance


__all__ = [
    "Client",
    "ClientError",
    "ErrorResponse",
    "GeneralError",
    "Instance",
    "UnexpectedResponse",
]



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in the additional IP address API methods. in BlueCat Address Manager."""
from ._enum import StrEnum


class AdditionalIPServiceType(StrEnum):
    """Additional IP Service Type"""

    LOOPBACK = "loopback"
    SERVICE = "service"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Values defined for SNMP Versions for SNMP Services in BlueCat Address Manager.
"""
from ._enum import StrEnum


class SNMPVersion(StrEnum):
    """Constants defined for SNMP versions."""

    V1 = "v1"
    V2C = "v2c"
    V3 = "v3"



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
Wrappers for standard dictionary REST results to make them mimic `suds`
dictionary results.
"""
from copy import deepcopy


class RESTDict(dict):
    """
    Wrapper for standard dictionary REST results to make them mimic `suds`
    dictionary results.
    """

    def __getattr__(self, item):
        return self[item]

    def __getstate__(self):
        pass

    def __setattr__(self, key, value):
        self[key] = value

    def __deepcopy__(self, memo):
        cls = self.__class__
        res = deepcopy(super(RESTDict, self))  # pylint: disable=super-with-arguments; legacy code
        return cls(res)



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values for types of the option in BlueCat Address Manager."""
from ._enum import StrEnum


class OptionType(StrEnum):
    """Constants used in the option types."""

    DHCP_CLIENT = "DHCPClient"
    DHCP_SERVICE = "DHCPService"
    DHCP_VENDOR_CLIENT = "DHCPVendorClient"
    DHCP6_CLIENT = "DHCP6Client"
    DHCP6_SERVICE = "DHCP6Service"
    DHCP6_VENDOR_CLIENT = "DHCP6VendorClient"
    DHCP_RAW = "DHCP_RAW"
    DHCPV6_RAW = "DHCPV6_RAW"
    DNS = "DNS"
    DNS_RAW = "DNS_RAW"



# Copyright 2020-2022 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Definition of a class for comparing versions."""
from typing import Union


class Version:
    """Version class."""

    def __new__(cls, ver: Union["Version", str]) -> "Version":
        """String input must be a dot-separated list of integers (e.g. '1.2.3')."""
        if isinstance(ver, Version):
            return ver
        if isinstance(ver, str):
            o = object.__new__(cls)
            o.__v = tuple(int(x) for x in ver.split("."))
            return o
        raise TypeError(
            f"Invalid argument type for {Version.__name__}(): {repr(type(ver).__name__)}"
        )

    def __str__(self) -> str:
        return ".".join(str(x) for x in self.__v)

    def __repr__(self) -> str:
        return f"{Version.__name__}({repr(str(self))})"

    def __cmp_helper(self, op: str, other: Union["Version", str]) -> bool:
        try:
            other = Version(other)
        except TypeError:
            return NotImplemented
        self, other = self.__v, other.__v
        if len(self) < len(other):
            self += (0,) * (len(other) - len(self))
        else:
            other += (0,) * (len(self) - len(other))
        return getattr(self, op)(other)

    def __eq__(self, other: Union["Version", str]) -> bool:
        return self.__cmp_helper("__eq__", other)

    def __ne__(self, other: Union["Version", str]) -> bool:
        return self.__cmp_helper("__ne__", other)

    def __lt__(self, other: Union["Version", str]) -> bool:
        return self.__cmp_helper("__lt__", other)

    def __le__(self, other: Union["Version", str]) -> bool:
        return self.__cmp_helper("__le__", other)

    def __gt__(self, other: Union["Version", str]) -> bool:
        return self.__cmp_helper("__gt__", other)

    def __ge__(self, other: Union["Version", str]) -> bool:
        return self.__cmp_helper("__ge__", other)



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Setup script for bluecat-libraries."""

from pathlib import Path
from setuptools import setup, find_packages

# IMPORTANT:
# When installing as editable, it doesn't matter what the current directory is when invoking
# `pip install -e <path-to-bclib>`, but when building a package, current directory has to be `HERE`.
HERE = Path(__file__).resolve().parent


def get_version() -> str:
    """Self-explanatory."""
    fp = HERE / "version.txt"
    if fp.exists():
        v = fp.read_text().strip()
    else:
        v = "0.0.0+placeholder"
    # print(f"The version for the package is going to be: {v}")
    return v


setup(
    name="bluecat-libraries",
    version=get_version(),
    author="BlueCat",
    maintainer="BlueCat",
    url="https://docs.bluecatnetworks.com",
    description="Modules for working with BlueCat products.",
    long_description=(HERE / "README.rst").read_text(),
    long_description_content_type="text/x-rst",
    license="http://www.apache.org/licenses/LICENSE-2.0",
    python_requires=">=3.9",
    package_dir={"": "src"},  # does not accept abs path
    packages=find_packages(where="src"),
    install_requires=[
        "requests>=2.26",
    ],
    extras_require={
        "dev": [
            "pytest>=6.2.5",
            "responses>=0.14.0",
            "paramiko>=2.7.2",
        ],
        "internal-suds": [
            "suds-py3>=1.4.4.1",
        ],
    },
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.9",
    ],
)



# Copyright 2020 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""
REST API Specific exceptions.
"""
from suds import WebFault
from bluecat_libraries.http_client import ErrorResponse


class Fault:
    """
    Wrapper for fault messages.
    """

    def __init__(self, fault):
        self.faultstring = "Server raised fault: %s" % fault


class RESTFault(WebFault, ErrorResponse):
    """
    Exception wrapper for REST that mimics suds' WebFault exception.
    """

    def __init__(self, fault, document=None):
        # Expected types:
        # * fault is str
        # * document is requests.Response
        # NOTE: Upcoming rename of parameters (will happen when use of RESTFault is replaced by
        # direct use of ErrorResponse):
        # * fault -> message
        # * document -> response
        f = Fault(fault)
        WebFault.__init__(self, f, document)
        ErrorResponse.__init__(self, f.faultstring, document)

    def __str__(self):
        return str(self.message)



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in the IPv4 IP group method."""
from ._enum import StrEnum


class IPGroupRangePosition(StrEnum):
    """Constants used to specify the position of the IP group range in the parent network."""

    END_OFFSET = "END_OFFSET"
    START_ADDRESS = "START_ADDRESS"
    START_OFFSET = "START_OFFSET"



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used by data collection probes in BlueCat Address Manager."""
from enum import IntEnum

from ._enum import StrEnum


class DefinedProbe(StrEnum):
    """Values of pre-defined SQL queries that have been triggered to collect data."""

    LEASE_COUNT_PER_DATE = "LEASE_COUNT_PER_DATE"
    NETWORK_BLOOM = "NETWORK_BLOOM"


class DefinedProbeStatus(IntEnum):
    """Values of status codes for the data collection probe"""

    INIT = 0
    INQUEUE = 1
    PROCESSING = 2
    COMPLETED = 3



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Enumeration class to be used in constant classes in BlueCat Address Manager."""
from enum import Enum


class StrEnum(str, Enum):
    """Enum where members are also (and must be) strings"""

    def __str__(self):
        return str(self.value)



# Copyright 2021 BlueCat Networks (USA) Inc. and its affiliates.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
"""Values used in GET /v1/getServerDeploymentStatus method in BlueCat Address Manager."""
from enum import IntEnum

from ._enum import StrEnum


class ServersDeploymentStatus(IntEnum):
    """Values of status codes for deployment of a particular server in BlueCat Address Manager."""

    EXECUTING = -1
    INITIALIZING = 0
    QUEUED = 1
    CANCELLED = 2
    FAILED = 3
    NOT_DEPLOYED = 4
    WARNING = 5
    INVALID = 6
    DONE = 7
    NO_RECENT_DEPLOYMENT = 8


class DeploymentTaskStatus(StrEnum):
    """
    Values of the overall deployment status and the deployment status of individual entities for deployment of
    a particular server in BlueCat Address Manager.
    """

    QUEUED = "QUEUED"
    STARTED = "STARTED"
    FINISHED = "FINISHED"
    TASK_NOT_FOUND = "TASK_NOT_FOUND"
